#!/usr/bin/python
# -*- coding: UTF-8 -*-
# 21.05.2017
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from cl_Entry import Entry
from cl_Combo import Combo
import database as db


# Hier wird ein Buchungsvorgang angelegt bzw. eine vorhandene Buchung geändert.
# Die Eingabefelder für Datum, Brutto, Netto und USt. werden geprüft auf zulässige Zeichen und Wertebereiche


# Der Index der Entryfelder - (nicht der Datenbank !)
DATE = 0
BRUTTO = 1
NETTO = 2
UST = 3
TEXT1 = 4
TEXT2 = 5


class Buchung(tk.Toplevel):
    # hier werden die zuletzt benutzten Konten gespeichert und beim nächsten Aufruf wieder ausgewählt
    konto_einnahmen = 0
    konto_ausgaben = 0

    # Dialog für eine neue Buchung - oder zum Ändern einer bestehenden Buchung,
    # wenn via 'key' der Primary Key eines bestehenden Datensatzes übergeben wird.
    def __init__(self, parent, key=None, callback=None):
        tk.Toplevel.__init__(self)
        self.resizable(width=False, height=False)
        self.bind("<Escape>", self.cancel)
        self.protocol("WM_DELETE_WINDOW", self.cancel)
        self.title("Buchung: {}".format(db.label))
        self.geometry("+{}+{}".format(parent.winfo_rootx()+75, parent.winfo_rooty()+75))
        self.key = key
        self.backup = False
        self.callback = callback
        self.table = db.table

        # die Standardumsatzsteuersätze einlesen für die Beschriftung der Radiobuttons
        self.ust_satz_2 = db.ustsaetze[2]
        self.ust_satz_1 = db.ustsaetze[1]

        self.combo = Combo(self, column=2, row=0, combolist=db.konten, height=18, callback=self.combo_changed)

        # das zuletzt benutzte Konto wird in der Combobox angewählt, differenziert nach Ausgaben und Eingaben
        if self.table == "ausgaben":
            idx = self.__class__.konto_ausgaben
        else:
            idx = self.__class__.konto_einnahmen
        # 'idx' ist der Index auf die Comboliste
        self.combo.set_selection(idx)

        radios = (("{}% Umsatzsteuer".format(self.ust_satz_2), self.ust_satz_2, 3, 1),
                  ("{}% Umsatzsteuer".format(self.ust_satz_1), self.ust_satz_1, 3, 2),
                  ("ohne Umsatzsteuer", 0, 3, 3),
                  ("freie Eingabe", -1, 3, 4))

        # den Umsatzsteuersatz gemäß der Vorgabe der Kontendefinition wählen
        self.ustsatz_var = tk.IntVar()
        self.ustsatz_var.set(db.get_ustsatz_by_kontoname(db.konten[idx]))

        for r in radios:
            ttk.Radiobutton(
                self, text=r[0], variable=self.ustsatz_var, width=25, value=r[1], takefocus=0,
                command=self.change_ust_modus).grid(column=r[2], row=r[3], padx=5, pady=5, sticky="nsw")

        self.netto_eingabe_modus = tk.IntVar()
        self.netto_btn = ttk.Checkbutton(
            self, text="Betrag netto eingeben", command=self.change_eingabe_modus, variable=self.netto_eingabe_modus)
        self.netto_btn.grid(column=3, row=5, padx=5, pady=5, sticky="nsw")

        # Bezeichner, Datentyp, spalte, zeile, callback)
        entrys = (("Datum : ", "date", 0, 1, None),
                  ("Brutto : ", "float", 0, 2, self.brutto_gets_input),
                  ("Netto : ", "float", 0, 3, self.netto_gets_input),
                  ("Ust : ", "float", 0, 4, self.ust_gets_input),
                  ("Textzeile 1 : ", "text", 0, 5, None),
                  ("Textzeile 2 : ", "text", 0, 6, None))

        self.entry_list = []
        # Platz für 6 Einträge + Ustsatz + Kontoauswahl vorbereiten
        self.entry_backup = [''] * 8
        for e in entrys:
            self.entry_list.append(Entry(self, label=e[0], entry_type=e[1], column=e[2], row=e[3], callback=e[4]))

        # es gibt im Moment noch keinen Datensatz, der zurückkopiert werden kann
        self.backup = False

        self.entry_list[NETTO].disable()
        self.entry_list[NETTO].set('0.00')
        self.entry_list[UST].disable()
        self.entry_list[UST].set('0.00')

        btn1 = ttk.Button(self, text="übernehmen", command=self.buchen)
        btn1.grid(column=3, row=6, padx=5, pady=5, sticky="nsew")
        btn1.bind('<Return>', self.btn_buchen_pressed)

        btn2 = ttk.Button(self, text="beenden", command=self.cancel)
        btn2.grid(column=3, row=7, padx=5, pady=5, sticky="nsew")
        btn2.bind('<Return>', self.btn_beenden_pressed)

        # Je nach dem ob eine neue Buchung angelegt oder eine bestehnde geändert werden soll,
        # hat der letzte Button unterschiedliche Funktion
        if self.key:
            self.btn = ttk.Button(self, text="Datensatz löschen", command=self.delete)
        else:
            self.btn = ttk.Button(self, text="letzte Buchung einfügen", state=tk.DISABLED, command=self.copy_values)
        self.btn.grid(column=3, row=8, padx=5, pady=5, sticky="nsew")

        # self.focus_set()           # focus auf das Fenster
        self.entry_list[0].setfocus()
        self.grab_set()            # alle Events werden nur nach w durchgereicht,
        #                          # das drunterliegende Fenster bleibt inaktiv, wird aber aktualisiert
        self.transient(parent)     # das Fenster bleibt oberhalb von Top

        # wenn eine Buchung geändert werden soll, dann hier die alten Werte einlesen und in die Eingabefelder übertragen
        if self.key:
            self.read_values(self.key)

        self.wait_window(self)

    def btn_buchen_pressed(self, e):
        self.buchen()

    def btn_beenden_pressed(self, e):
        self.cancel()

    def combo_changed(self, e):
        """
        Wenn ein neues Konto ausgewählt wurde, dann wird die Umsatzsteuervorgabe für dieses Konto gewählt
        und der zugehörige Radiobutton aktiviert.
        :param e: (Auswahl, Kontobezeichnung)
        :return:
        """
        self.ustsatz_var.set(db.get_ustsatz_by_kontoname(e[1]))
        # die Eingabefelder entsprechend der Kontovorgaben einstellen:
        self.change_ust_modus()

    def change_eingabe_modus(self):
        # Der Eingabemodus (Brutto / Netto) wird geändert.
        # Es dient entweder das Eingabefeld 'Brutto' oder 'Netto' als Grundlage für die Berechnung
        # der beiden anderen Werte.
        # Beim Umstellen des Modus werden zunächst nur die entsprechenden Eingabefelder (de-) aktiviert
        # Als Eingabemodus ist netto gewählt
        if self.netto_eingabe_modus.get():
            self.entry_list[BRUTTO].disable()
            self.entry_list[NETTO].enable()
        else:
            self.entry_list[BRUTTO].enable()
            self.entry_list[NETTO].disable()

    def delete(self):
        # Wird nur beim Ändern von vorhandenen Buchungen benutzt.
        # Es wird die Abfrage ausgeführt, ob der Datensatz gelöscht werden soll
        # Wenn ja, dann wird der Primary Key des Datensatzes zurückgegeben.
        # Das aufrufende Programm muss unterscheiden, ob ein geänderter Datensatz zurückgeliefert wird
        # oder - wie beim Löschen - nur ein Tuple mit dem Primary Key
        msg = "- Key: {}\n- Datum: {}\n- Brutto: {}\n- Netto {}\n- Ust: {}\n- Text1: {}\n- Text2: {}".format(
                self.key,
                self.entry_list[DATE].get(),
                self.entry_list[BRUTTO].get(),
                self.entry_list[NETTO].get(),
                self.entry_list[UST].get(),
                self.entry_list[TEXT1].get(),
                self.entry_list[TEXT2].get())
        if messagebox.askyesno("Datensatz löschen", "Diesen Datensatz\n\n{}\n\nwirklich löschen ?".format(msg)):
            # Rückgabe: Tuple mit Primary Key, das Löschen selbst übernimmt das aufrufende Programm
            self.callback((self.key,))
            # Das Fenster schließen und zurück zum Treeview
            self.destroy()

    def read_values(self, key):
        # liest die aktuellen Werte eines Datensatzes aus der Datenbank ein
        # holt den ersten und einzigen Datenatz mit fetchone()
        # und überträgt die Werte in die Entryfelder
        rec = db.get_buchung(db.table, key).fetchone()

        # ID, konto, date, text1, text2, netto, ust, ustsatz
        # Datum umformatieren
        d = str(rec[2])
        self.entry_list[DATE].set("{}.{}.{}".format(d[4:], d[2:4], d[0:2]))
        self.entry_list[BRUTTO].set("{:.2f}".format((rec[5] + rec[6]) / 100))
        self.entry_list[NETTO].set("{:.2f}".format(rec[5] / 100))
        self.entry_list[UST].set("{:.2f}".format(rec[6] / 100))
        self.entry_list[TEXT1].set(rec[3])
        self.entry_list[TEXT2].set(rec[4])
        # den Umsatzsteuersatz einlesen und den richtigen Radiobutton aktivieren
        self.ustsatz_var.set(rec[7])
        # wenn die Umsatzsteuer als freie Eingabe definiert ist, dann die Felder UST und Netto enablen
        # und den Checkbutton für Eingabe netto setzen
        if rec[7] < 0:
            self.entry_list[BRUTTO].set("")
            self.entry_list[BRUTTO].disable()
            self.entry_list[NETTO].enable()
            self.entry_list[UST].enable()
            self.netto_eingabe_modus.set(1)
        # die Combobox mit den Kontenbezeichnungen einstellen
        # geliefert ist die interne Kontenbezeichnung in rec[1]
        # die Kontobezeichnung wird über db.get_name(key) ermittelt
        self.combo.set_name(db.get_name(rec[1]))

    def change_ust_modus(self):
        # Wird ausgeführt, wenn der Radiobutton Umsatzsteuersatz geändert wird,
        # wenn ein anderes Konto ausgewählt wird
        # (dabei kann ein anderer Umsatzsteuersatz als Vorgabe eingestellt werden)
        # oder wenn der letzte Datensatz zurückkopiert wird
        # (hier soll sichergestellt werden, dass auch die Eingabefelder wie vorher eingestellt sind).

        # wenn beim Umsatzsteuersatz 'freie Eingabe' gewählt ist
        if self.ustsatz_var.get() < 0:
            self.entry_list[BRUTTO].disable()
            self.entry_list[NETTO].enable()
            self.entry_list[UST].enable()
            # Das Feld brutto löschen, es wird nicht benötigt
            self.entry_list[BRUTTO].set("")
            # Netto für Eingabe-Modus "freie Eingabe" aktivieren
            # self.netto_eingabe_modus.set(1)
            # und den Checkbutton sperren. solange die Freie Eingabe gewählt ist
            self.netto_btn.config(state=tk.DISABLED)
            return

        # wenn ein Umsatzsteuersatz 0, ust_1, ust_2 gewählt ist
        else:
            # falls vorher dieser Button disabled war
            self.netto_btn.config(state=tk.NORMAL)
            # Die Eingabefelder dis/enablen
            # wenn 'Netto' als Eingabemodus gewählt ist
            if self.netto_eingabe_modus.get():
                self.entry_list[BRUTTO].disable()
                self.entry_list[NETTO].enable()
            # wenn 'Brutto' als Eingabemodus gewählt ist
            else:
                self.entry_list[BRUTTO].enable()
                self.entry_list[NETTO].disable()
            # das Umsatzsteuerfeld ist immer disabled
            self.entry_list[UST].disable()
        # Und die Umsatzsteuer neu ausrechnen, basiernd auf Netto bzw. Brutto
        # sowie dem gewählten Umsatzsteuersatz
        # Bei Eingabe via Brutto-Feld
        if not self.netto_eingabe_modus.get():
            self.brutto_gets_input(self.entry_list[BRUTTO].get())
        # Bei Eingabe via Netto-Feld
        else:
            self.netto_gets_input(self.entry_list[NETTO].get())

    def brutto_gets_input(self, brutto):
        # wird ausgeführt, wenn eine Eingabe im Brutto-Feld ausgeführt wurde
        # dann wird der zugehörige Wert zu Netto und Umsatzsteuer gerechnet und ausgegeben
        # print("brutto_get_input", brutto)
        brutto = float(self.komma2punkt(brutto))
        # Wenn die Umsatzsteuer berechnet werden soll
        if self.ustsatz_var.get() >= 0:
            netto = brutto * 100 / (float(self.ustsatz_var.get() + 100))  # float(self.entry_list[UST].get())
            self.entry_list[NETTO].set("{:.2f}".format(netto))
            self.entry_list[UST].set("{:.2f}".format(brutto - netto))

    def netto_gets_input(self, netto):
        # wird ausgeführt, wenn eine Eingabe im Netto-Feld ausgeführt wurde.
        # dann wird der zugehörige Wert für Brutto und Umsatzsteuer gerechnet und ausgegeben
        # Wenn die Umsatzsteuer berechnet werden soll
        netto = float(self.komma2punkt(netto))
        if self.ustsatz_var.get() >= 0:
            brutto = netto * (100.0 + float(self.ustsatz_var.get())) / 100
            self.entry_list[BRUTTO].set("{:.2f}".format(brutto))
            self.entry_list[UST].set("{:.2f}".format(brutto - netto))

    def ust_gets_input(self, ust):
        # wird ausgeführt, wenn eine Eingabe im USt-Feld ausgeführt wurde.
        # das ist nur in der freien Eingabe möglich
        # dann wird Umsatzsteuer + Netto in Brutto geschrieben (Brutto ist disabled)
        ust = float(self.komma2punkt(ust))

        if not self.netto_eingabe_modus.get():
            netto = float(self.komma2punkt(self.entry_list[NETTO].get()))
            brutto = netto + ust
            print(brutto, netto, ust)
            self.entry_list[BRUTTO].set("{:.2f}".format(brutto))

    def cancel(self, e=None):
        # Die Einstellungen für das zuletzt genutzte Konto speichern
        # es wird der Index in der Combolist gespeichert.
        if self.table == "einnahmen":
            Buchung.konto_einnahmen = self.combo.get_selection()[0]
        else:
            Buchung.konto_ausgaben = self.combo.get_selection()[0]
        self.destroy()

    def buchen(self):
        # wird ausgeführt, wenn der Button "übernehmen" gedrückt wurde

        # Die Datumseingabe auf Vollständigkeit und Gültigkeit prüfen
        date = chkdate(self.entry_list[DATE].get())
        if date is None:
            return
        else:
            date, datestring = date
            # Falls der Eintrag später kopiert werden soll hier schon einmal eine Kopie in der Backupliste anlegen
            self.entry_backup[0] = datestring

        # Brutto, Netto und Ust einlesen, Dezimalkomma gegen Dezimalpunkt tauschen und zu Float machen
        brutto = float(self.komma2punkt(self.entry_list[BRUTTO].get()))
        netto = float(self.komma2punkt(self.entry_list[NETTO].get()))
        ust = float(self.komma2punkt(self.entry_list[UST].get()))

        # es kann Rundungsfehler geben bei "int(netto * 100)", daher später "int(round(netto * 100, 2)"
        # print("buchen() ", brutto, netto, ust, netto * 100, int(netto * 100))

        # Prüfen, ob eine Eingabe vorliegt
        # if netto < 0.01 and netto > -0.01:
        if -0.01 < netto < 0.01:
            if not messagebox.askyesno(
                    "Prüfung der Eingabe", "Der Buchungsbetrag ist '0.00' !\nTrotzdem übernehmen ?"):
                return

        # Prüfen, ob ein Texteintrag in Textzeile 1 vorhanden ist
        if not len(self.entry_list[TEXT1].get()):
            messagebox.showerror("Prüfung Textzeile 1", "Kein Buchungstext in Textzeile 1 !")
            return

        # Das " und ' in der Textzeile führt beim Schreiben in die Datenbank zu einem Fehler
        if '"' in self.entry_list[TEXT1].get():
            messagebox.showerror("Prüfung Textzeile 1", 'Unzulässiges Zeichen [ " ] in Textzeile 1')
            return

        if '"' in self.entry_list[TEXT2].get():
            messagebox.showerror("Prüfung Textzeile 2", 'Unzulässiges Zeichen [ " ] in Textzeile 2')
            return

        if "'" in self.entry_list[TEXT1].get():
            messagebox.showerror("Prüfung Textzeile 1", "Unzulässiges Zeichen [ ' ] in Textzeile 1")
            return

        if "'" in self.entry_list[TEXT2].get():
            messagebox.showerror("Prüfung Textzeile 2", "Unzulässiges Zeichen [ ' ] in Textzeile 2")
            return

        # Das Konto als Textelement aus der Combobox holen und die zugehörige KontoId bestimmen
        tmp = self.combo.get_selection()
        konto = db.get_key(tmp[1])

        # Sonderfall KFZ-Betriebskosten = Konto 90
        # hier kann ist eine verkürzte Eingabe "Kilometer, Liter" möglich,
        # zwischen beiden muss ein Komma stehen (der Punkt ist das Dezimaltrennzeichen)
        # die Formatierung wird automatisch übernommen
        # Format 35666 [km], 35.00 [L]
        if konto == 90:
            # Bemüht sich um die Formatierung, wenn die Prüfung ohne Erfolg ist, dann wird die Eingabe
            # unbeanstandet übernommen
            self.format_betriebskosten()

        # ein Callback wird beim Ändern genutzt, um Änderungen zurück in den Treeview zu schreiben
        if self.callback:
            # zuerst den Eintrag in die Datenbank zurückschreiben
            db.update_buchung(
                self.table, self.key, konto, date,
                self.entry_list[TEXT1].get(),
                self.entry_list[TEXT2].get(),
                int(round(netto * 100, 2)),
                int(round(ust * 100, 2)),
                int(self.ustsatz_var.get()))

            # diese Funktion liefert den Inhalt des geänderten Datensatzes zurück,
            # damit die Daten im Treeview aktualisiert werden können
            self.callback((
                self.key, konto, datestring,
                self.entry_list[TEXT1].get(),
                self.entry_list[TEXT2].get(),
                brutto, netto, ust))

            # Nach dem Ändern eines Datensatzes das zugehörige Fenster schließen
            self.destroy()

        # einen neuen Eintrag in der Datenbank ablegen
        else:
            db.add_buchung(
                self.table, konto, date,
                self.entry_list[TEXT1].get(),
                self.entry_list[TEXT2].get(),
                int(round(netto * 100, 2)),
                int(round(ust * 100, 2)),
                int(self.ustsatz_var.get()))
            # alle Eingabewerte zwischenspeichern, sie können dann per Button wieder in die EingabeMaske kopiert werden
            # für eine neue Eingabe.
            # Das Datum ist bereits oben nach der Prüfung gesichert worden.
            # self.entry_backup[0] = self.entry_list[0].get()
            self.entry_backup[1] = "{:.2f}".format(brutto)
            self.entry_backup[2] = "{:.2f}".format(netto)
            self.entry_backup[3] = "{:.2f}".format(ust)
            for i in range(4, len(self.entry_list)):
                self.entry_backup[i] = self.entry_list[i].get()
            self.entry_backup[i + 1] = self.ustsatz_var.get()
            self.entry_backup[i + 2] = self.combo.get_selection()[0]

            # nach dem ersten Backup den Button zum Kopieren enablen
            if not self.backup:
                self.backup = True
                self.btn.configure(state=tk.NORMAL)

            # alle Eingabefelder löschen
            self.entry_list[DATE].set("")
            self.entry_list[BRUTTO].set("")
            self.entry_list[NETTO].set("")
            self.entry_list[UST].set("")
            self.entry_list[TEXT1].set("")
            self.entry_list[TEXT2].set("")
            # den Focus auf das erste Feld mit dem Datum setzen
            self.entry_list[DATE].setfocus()

    @staticmethod
    def komma2punkt(val):
        """
        Wenn der Eingabestring ein DezimalKomma enthält, dann wird es gegen einen Punkt ausgetauscht
        :param val:
        :return:
        """
        if "," in val:
            val = val.replace(",", ".")
        return val

    def format_betriebskosten(self):
        """
        Die zweite Eingabezeile für KFZ-Betriebskosten in ein definiertes Format bringen
        Eingabe: Kilometerstand [km], Menge [L]
        Dazu die Eingabe per split(',') aufteilen: es müssen genau 2 Elemente vorhanden sein.

        """
        i = self.entry_list[TEXT2].get()
        i = i.split(',')
        if not len(i) == 2:
            return False

        km = i[0].strip()
        if km.isnumeric():
            km = int(km)
        else:
            return False
        li = i[1].strip()
        try:
            li = round(float(li), 2)
        except:
            return False
        # Zeile formatieren zu "123456 km, 12.34 L"
        self.entry_list[TEXT2].set("{:6} km, {:5.2f} L".format(km, li))
        return True

    def copy_values(self):
        # Kopiert die Werte der vorausgegangenen Buchung zurück in die Eingabefelder
        # und setzt den Focus auf den Datumseintrag
        for i in range(len(self.entry_list)):
            self.entry_list[i].set(self.entry_backup[i])

        # Sicherheitshalber die Umsatzsteuer- und Kontozuordnung des letzten Datensatzes übernehmen
        # Die könnten vor dem Kopieren manuell geändert worden sein.
        self.ustsatz_var.set(self.entry_backup[i + 1])
        self.combo.set_selection(self.entry_backup[i + 2])

        # die Eingabefelder entsprechend der gewählten Voreinstellungen de- /aktivieren
        self.change_ust_modus()

        self.entry_list[DATE].setfocus()


def chkdate(datum):
    """
    Prüft die Eingabe im Eingabefeld 'Datum' auf Vollständigkeit und kalendarische Gültigkeit
    :param datum:
    :return: None oder Datum im Format "YYMMDD" als INT + "DD.MM.YY" als String
    """
    label = "Prüfung des Datums"
    chk_year = int(db.dict_config.get("buchungsjahr"))

    d = datum.split('.')
    if len(d) < 2:
        messagebox.showerror(label, "Die Datumseingabe ist nicht vollständig: '{}'".format(datum))
        return None

    elif len(d) == 2:
        if len(d[1]):
            year = chk_year

        else:
            messagebox.showerror(label, "Es fehlt eine Monatsangabe: '{}'".format(datum))
            return None

    elif len(d) == 3:
        if len(d[2]):
            year = int(d[2])
            if not year == chk_year:
                messagebox.showerror(label, "Falsche Jahresangabe: '{}'\nBuchungen nur für Jahr 20{} zulässig !".format(
                                        year, chk_year))
                return None
        else:
            year = int(db.dict_config.get("buchungsjahr"))

    # year += 2000
    day = int(d[0])
    month = int(d[1])
    if lastday_of_month(month, year) < day:
        messagebox.showerror(label, "Die Datumseingabe ist ungültig: {}".format(datum))
        return None

    date_int = year * 10000 + month * 100 + day
    date_str = "{:02}.{:02}.{:02}".format(day, month, year)

    return date_int, date_str


def lastday_of_month(month, year):
    """
    :param month:
    :param year:
    :return: gibt zurück, wieviele Tage der Monat hat
    """
    days_of_month = 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31

    if not month == 2:
        return days_of_month[month - 1]
    else:
        if Date_Is_Leap(year):
            return 29
        else:
            return 28


def Date_Is_Leap(y):
    """ Liefert zurück, ob das Jahr 'y' ein Schaltjahr ist
    :param y:
    """
    if ((y % 4 == 0) & (y % 100 != 0)) | (y % 400 == 0):
        return True
    else:
        return False

if __name__ == "__main__":
    top = tk.Tk()
    db.open_db()
    db.init_konten()
    db.define_table("ausgaben")
    Buchung(top)
    db.close_db()
    top.destroy()
