#!/usr/bin/python
# -*- coding: UTF-8 -*-
# 08.05.2016

import tkinter as tk
from tkinter import ttk
import database as db


class Combo(object):
    def __init__(self, master=None, column=0, row=0, height=12, width=20, callback=None, combolist=None):

        self.combolist = combolist
        self.last_selection = 0
        self.master = master
        self.cb = callback

        self.combo = ttk.Combobox(self.master, values=self.combolist,
                                  state="readonly", takefocus=0, height=height, width=width)
        self.combo.set(self.combolist[self.last_selection])
        self.combo.bind("<<ComboboxSelected>>", self._change_selection)
        self.combo.grid(column=column, columnspan=4, row=row, padx=5, pady=5, sticky="nswe")

    def _change_selection(self, event):
        """
        Wird bei jeder Änderung aufgerufen und schreibt die aktuelle Auswahl in die Variable last_selection
        Ist eine Callback-Funktion definiert, dann wird die ausgeführt und liefert (Idx, Bezeichner) zurück.
        """
        self.last_selection = event.widget.current()
        if self.cb:
            self.cb((self.last_selection, self.combolist[self.last_selection]))

    def get_selection(self):
        """
        Gibt den Index und den Bezeichner der zuletzt erfolgten Auswahl zurück
        :return: (idx, name)
        """
        return self.last_selection, self.combolist[self.last_selection]

    def set_selection(self, idx):
        """
        Der Index der gewünschten Selektion wird übergeben, das zugehörige Element wird angezeigt
        :param idx: Index der zu aktivierenden Auswahl in der Combobox
        :return:
        """
        """ Der Index der Auswahl ist bekannt """
        self.combo.set(self.combolist[idx])
        self.last_selection = idx

    def set_name(self, name):
        """
        Die Bezeichnung der gewünschten Selektion wird übergeben, das zugehörige Element wird angezeigt
        :param name: Bezeichnung der Auswahl in der Combobox
        """
        self.combo.set(name)
        self.last_selection = db.get_index(name)

    def disable(self):
        self.combo.configure(state = tk.DISABLED)

    def update(self, pos, newlist):
        """
        Die Liste mit den Einträgen wird ausgetauscht, die neue Liste ist newlist
        Der Eintrag an Position 'pos' wird angezeigt.
        :param pos:  der Eintrag mit Index pos soll angezeigt werden
        :param newlist: neue Liste
        """
        self.combolist = newlist
        self.combo.configure(values=newlist)
        self.combo.set(newlist[pos])

if __name__ == "__main__":

    def callback(data):
        print("Callback: gewählt wurde", data)

    table = ("Option 1", "Option 2", "Option 3")

    top = tk.Tk()
    frm = tk.Frame(top)
    frm.grid(column=0, row=0)

    Combo(master=frm, height=12, combolist=table, callback=callback)
    top.title("ttk Combobox")
    top.mainloop()

