#!/usr/bin/python
# -*- coding: UTF-8 -*-
# 07.05.2017

import tkinter as Tk
from tkinter import ttk


class Entry(object):
    """
    Ein Label plus ein Entry-Widget
    Die Eingabe kann auf zulässige Werte geprüft werden, unzulässige Eingaben werden abgewiesen
    """
    # Ein Callback, das bei "RETURN" aufgerufen wird
    cb_return = None

    def __init__(self, master, label=None, column=0, columnspan=1, row=0, entry_type="float", callback=None):
        Tk.Label(master, text=label, width=10, anchor='e').grid(
                column=column, columnspan=columnspan, row=row, padx=5, pady=5, sticky="nse")
        self.__var = Tk.StringVar()
        self.cb = callback
        self.master = master

        justify = "right"
        if entry_type == "float":
            width = 10
        elif entry_type == "text":
            width = 25
            justify = "left"
        elif entry_type == "date":
            width = 8
        else:
            print("Falscher entrytype")
            exit()
        self.entry_type = entry_type

        self.__entry = ttk.Entry(
                master,                     # das Parent-Window
                # state="normal",           #
                textvariable=self.__var,    # die steuernde Variable
                justify=justify,
                font=("Courier", "10", "bold"),
                width=width                   # die Breite des Eingabefeldes
                )

        self.__entry.grid(column=column + 1, row=row, padx=5, pady=5, sticky="nsw")

        if entry_type == "date":
            self.__check = self.__entry.register(self.is_date)
            self.__entry.configure(validate='key', validatecommand=(self.__check, '%P'))

        elif entry_type == "float":
            self.__check = self.__entry.register(self.is_float)
            self.__entry.configure(validate='key', validatecommand=(self.__check, '%P'))

        elif entry_type == "text":
            self.__check = self.__entry.register(self.textlen)
            self.__entry.configure(validate='key', validatecommand=(self.__check, '%P'))

        self.__entry.bind('<Key-Return>', self.__return)

    def __return(self, event):
        t = self.__entry.tk_focusNext()
        t.focus_set()

    def set(self, value):
        self.__var.set(value)

    def setfocus(self):
        self.__entry.focus_set()

    def enable(self):
        self.__entry.config(state="normal")

    def disable(self):
        self.__entry.config(state="disabled")

    def get(self):
        """ Die Variable kann auch leer sein ! """

        i = self.__var.get()
        if not i and self.entry_type == "float":
            i = '0'
        return i

    def is_date(self, new_value):
        d = new_value.split('.')

        for i in range(len(d)):
            if len(d[i]) and not d[i].isdigit():
                self.__entry.bell()
                return False

            if i == 0 and len(d[0]):
                if len(d[0]) == 1 and int(d[0]) == 0:
                    return True
                elif int(d[0]) < 1 or int(d[0]) > 31:
                    # print(d[0])
                    self.__entry.bell()
                    return False

            elif i == 1 and len(d[1]):
                if len(d[1]) == 1 and int(d[1]) == 0:
                    return True
                elif int(d[1]) < 1 or int(d[1]) > 12:
                    # print(d[1])
                    self.__entry.bell()
                    return False

            elif i == 2 and len(d[2]):
                if len(d[2]) == 1 and int(d[2]) == 0:
                    return True
                if int(d[2]) < 0 or int(d[2]) > 99:
                    self.__entry.bell()
                    return False
        return True

    def is_float(self, new_value):
        """ Achtung: self.__var.get() liefert den unveränderten Wert zurück ! """

        # print("is_float", new_value)

        if not new_value:            # Eine leeren Eingabe akzeptieren
            if self.cb:
                self.cb("0")
            return True

        if new_value == "-":        # Negative Werte zulassen (genaugenommen hier: '-' als '-0')
            return True

        # Prüfung der Eingabe auf Float mit maximal 2 Nachkommastellen
        # Dabei darf auch ein Komma als Dezimaltrennzeichen verwendet werden.
        try:
            # Falls ein Komma als Dezimaltrennzeichen verwendet wird
            if ',' in new_value:
                new_value = new_value.replace(",", ".")
            #
            if '.' in new_value:
                v = new_value.split('.')

                # Falls genau 1 Dezimaltrenner vorhanden war, ist len(v) == 2
                # dann dürfen aber nicht mehr als 2 Nachkommastellen eingegeben werden
                if len(v) == 2 and len(v[1]) > 2:
                    return False
                # Falls len(v) > 2, dann waren mindestens 2 Dzimalpunkte vorhanden
                if len(v) > 2:
                    return False

            # jetzt wird die Typumwandlung nach float versucht
            # das Ergebnis ist ohne Bedeutung, lediglich ein Fehler zählt
            float(new_value)

        except ValueError:
            # print("is_float_check", new_value, v)
            self.__entry.bell()
            return False
        if self.cb:
            self.cb(new_value)
        return True

    def textlen(self, new_value):
        """ Achtung: self.__var.get() liefert den unveränderten Wert zurück ! """

        if len(new_value) < 25:            # Auf maximale Länge testen
            return True
        else:
            self.__entry.bell()
            return False
