#!/usr/bin/python
# -*- coding: UTF-8 -*-
# 08.07.2021

# Bei Update '' gege  "" getauscht, ' im Test ist nun
# auch beim Update zulässig

# In diesem Modul wird die Definition von Klassen verzichtet,
# dadurch wurden leider einige globals notwendig.


import sqlite3 as sq
from tkinter import messagebox
from io import StringIO
import datetime
from readxml import xml_read
import locale as lc
lc.setlocale(lc.LC_ALL, "")

# Positionen in der Datenbank
KEY = 0
KONTO = 1
DATE = 2
TEXT1 = 3
TEXT2 = 4
NETTO = 5
UST = 6
USTSATZ = 7

timefilter = ""
kontofilter = ""
# Listen mit den Kontenbezeichnungen getrennt nach Ausgaben und Einnahmen
# Darauf wird über den Index (=Reihenfolge in der Combobox) zugegriffen
ausgabe_konten = []
einnahme_konten = []

# Dictionarys zur Suche des Kontonames über die KontoId und umgekehrt
# In dict_konten_by_key stehen die KontoInformationen (Key, Bezeichnung)
dict_konten_by_key = {}
dict_konten_by_name = {}
dict_ust_by_name = {}

# Dictionary für die Konfigurationsdaten (Buchungsjahr, Umsatzsteuersätze, Zeilenzahl pro Druckseite)
dict_config = {}

# globals
is_changed = False
db = None
cur = None
ustsaetze = [0, 0, 0]         # die Umsatzsteuersätze als Array

# wird von allen Modulen genutzt, um auf die richtige Tabelle bzw. deren Kontendefinitionen zugreifen zu können
# der Name der aktuellen Tabelle, der Text in der Fensterdkoration und die Liste der Kontenbezeichnungen
# für die Combobox bei der Eingabe
table = "ausgaben"
label = "Ausgaben"
konten = ausgabe_konten


def define_table(dbtable):
    # Hier wird die zu bearbeitende Tabelle festgelegt und damit:
    # der Name der Tabelle
    # der Ausgabetext im jeweiligen Fenster (label)
    # die zugehörige Kontoliste (konten)
    # Verschiedene Programmteile holen sich hier die aktuellen Einstellungen ab !
    global table, konten, label
    if dbtable == "ausgaben":
        table = dbtable
        konten = ausgabe_konten
        label = "Ausgaben"

    elif dbtable == "einnahmen":
        table = dbtable
        konten = einnahme_konten
        label = "Einnahmen"

    elif dbtable == "überschuss":
        label = "Überschuss"
        # Diese Kontenbezeichnung wird in der Comboboy ausgegeben, wenn der Überschuss ausgewertet werden soll
        konten = ('-----',)

    else:
        print("error setting dbtable")
        exit()


def init_konten():
    global ausgabe_konten, einnahme_konten, dict_konten_by_key, dict_konten_by_name, usttab, dict_config, \
            ustsaetze, dict_ust_by_name
    # alle Eintrag zurücksetzen, falls ein neues Konto eingelesen wird
    dict_konten_by_key = {}
    dict_konten_by_name = {}
    dict_ust_by_name = {}
    dict_config = {}
    usttab = []
    ausgabe_konten = []
    einnahme_konten = []

    # die Konten einlesen, Listen für die Comboboxen erstellen
    rows = cur.execute("""SELECT * FROM konten ORDER BY id""")
    for row in rows:
        dict_konten_by_key.setdefault(row[0], row[2])
        dict_konten_by_name.setdefault(row[2], row[0])
        dict_ust_by_name.setdefault(row[2], row[1])
        # alle Konten mit einer ID < 100 werden als Ausgabenkonten interpretiert
        # es werden hier zwei Listen erstellt
        # - für die Einnahmekonten
        # - für die Ausgabenkonten
        if row[0] < 100:
            ausgabe_konten.append(row[2])
        else:
            einnahme_konten.append(row[2])

    # die allgemeine Konfiguration
    rows = cur.execute("""SELECT * FROM config""")
    # alles ersteienmal in das dict_config schreiben
    for row in rows:
        dict_config.setdefault(row[0], row[1])

    # nun die gesuchten Daten aus dem Dictionary fischen
    ustsaetze[1] = dict_config.get("ust_1")
    ustsaetze[2] = dict_config.get("ust_2")
    # die Umsatzsteuervorgabe ist in den Konten nur als Index auf das Array ustsaetze abgelegt [0, 1, 2].
    # Hier die konkreten Umsatzsteuervorgaben holen
    for key in dict_ust_by_name:
        i = int(dict_ust_by_name.get(key))
        dict_ust_by_name[key] = ustsaetze[i]


def chk_database(file):
    # Prüft, ob die erwarteten Tabellen in der gewählten Datenbank vorhanden sind
    # bei Erfolg, wird True zurückgegeben, sonst False

    success = True
    try:
        _db = sq.connect(file)
        _cur = _db.cursor()
        row = _cur.execute("""SELECT * FROM ausgaben""").fetchone()
        # eine leere Tabelle liefert None zurück
        if row:
            if not len(row) == 8:
                success = False
        row = _cur.execute("""SELECT * FROM einnahmen""").fetchone()
        # eine leere Tabelle liefert None zurück
        if row:
            if not len(row) == 8:
                success = False
        row = _cur.execute("""SELECT * FROM konten""").fetchone()
        if not len(row) == 3:
            success = False
        row = _cur.execute("""SELECT * FROM config""").fetchone()
        if not len(row) == 2:
            success = False
    except:
        success = False

    finally:
        _db.close()
        return success


def open_db(name="./db/Beispiel_2015.db"):
    global db, cur
    # es muss vorher geprüft worden sein, ob die Datenbank die richtigen Tabellen enthält
    # -> chk_database()

    # falls bereits eine Datenbank geöffnet ist, dann diese zuerst schließen
    if db:
        db.commit()
        db.close()

    db = sq.connect(name)
    cur = db.cursor()


def close_db():
    global db, cur
    if cur:
        cur.close()
        cur = None
    if db:
        print("Database closed")
        db.close()
        db = None


def count_entrys():
    """
    Zählt die Anzahl der Einträge in den Tabellen Einnahmen und Ausgaben
    :return: Anzahl_Einträge in Einnahmen, Anzahl Einträge in Ausgaben
    """
    return ((cur.execute("""SELECT COUNT (*) FROM einnahmen""").fetchone())[0],
            (cur.execute("""SELECT COUNT (*) FROM ausgaben""").fetchone())[0])


def get_key(name):
    """
    holt zu einer Kontobezeichnung den zugehörigen Key/Id aus der Datenbank (nicht den Index in der Combobox)
    :param name:  Kontobezeichnung aus Combobox
    :return: id aus der Datenbank - oder -1 wenn der Eintrag nicht vorhanden ist
    """
    # Einnahmen und Ausgaben stehen hier gemeinsam im Dictionary
    return dict_konten_by_name.get(name, -1)


def get_name(key):
    """
    holt zu einem Schlüssel die zugehörige Kontobezeichnung
    :param key: = Konto_Id (der Datenbank) der KontoBezeichnung aus der Combobox
    :return: id aus der Datenbank - oder -1 wenn der Eintrag nicht vorhanden ist
    """
    return dict_konten_by_key.get(key, "unbekannt: {}".format(key))


def get_index(name):
    """
    Holt den Index einer Kontenbezeichnung in der Combobox , dabei wird entweder auf die Einträge der Einnahmen- oder
    der Ausgabenkonten zugegriffen.
    In 'konto' steht die Liste mit den aktuellen Kontenbezeichnungen (ausgaben oder einnahmen)
    :param name: Bezeichnung des Eintrages in der Combobox
    :return: Index des Eintrages in der Combobox, oder -1
    """
    for i, konto in enumerate(konten):
        if konto == name:
            return i
    return -1


def get_config(key, default=0):
    # Holt aus den Konfigruationdaten den zum Key passenden Integer-Wert
    return int(dict_config.get(key, default))


def get_ustsatz_by_kontoname(name, default=0):
    # Holt die Umsatzsteuervorgabe zum Konto 'name'
    return dict_ust_by_name.get(name, default)


def get_ustsatz_idx(ust):
    """
    Holt den Index für den Umsatzsteuersatz 'ust', so wie er in der XML-Datei verwendet wird:
    :param ust: Umsatzsteuersatz in Prozent
    :return:
    -> 0 = keine Umsatzsteuer
    -> 1 = ermäßigte Umsatzsteuer
    -> 2 = volle UmsatzsteuerIndex -
    oder -1 bei Fehler
    """
    for i, value in enumerate(ustsaetze):
        if value == ust:
            return i
    return -1


def add_tables():
    # fügt die Tabellen in eine neue Datenbank ein
    cur.execute("""CREATE TABLE IF NOT EXISTS ausgaben (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            konto INTEGER DEFAULT 0,
            date INTEGER DEFAULT 0,
            text1 VARCHAR(24) DEFAULT "",
            text2 VARCHAR(24) DEFAULT "",
            netto INTEGER DEFAULT 0,
            ust INTEGER DEFAULT 0,
            ustsatz INTEGER DEFAULT 0
            )""")

    cur.execute("""CREATE TABLE IF NOT EXISTS einnahmen (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            konto INTEGER DEFAULT 0,
            date INTEGER DEFAULT 0,
            text1 VARCHAR(24) DEFAULT "",
            text2 VARCHAR(24) DEFAULT "",
            netto INTEGER DEFAULT 0,
            ust INTEGER DEFAULT 0,
            ustsatz INTEGER DEFAULT 0
            )""")

    cur.execute("""CREATE TABLE IF NOT EXISTS konten (
            id INT UNIQUE,
            ustsatz INT,
            bezeichnung VARCHAR(24) UNIQUE
            )""")

    cur.execute("""CREATE TABLE IF NOT EXISTS config (
            key VARCHAR(20) UNIQUE,
            value VARCHAR(40)
            )""")


def make_database_from_xml(file):
    """
    Legt eine neue Datenbank an, dabei werden die Kontodefinitionen aus einer XML-Datei importiert.
    Die Datenbank wird am Ende wieder geschlossen.
    :param file: anzulegende Datenbank
    """
    open_db(file)
    add_tables()
    # Konfigurationsdaten und Konten als Listen aus der XML-Datei abholen
    configs, _konten = xml_read("config.xml")

    for konto in _konten:
        cur.execute("""INSERT INTO konten (id, ustsatz, bezeichnung) VALUES {}""".format(konto))

    for config in configs:
        cur.execute("""INSERT INTO config (key, value) VALUES {}""".format(config))

    db.commit()
    close_db()


def update_config(d_config, d_konten_name, d_konten_ust):
    """
    Nachdem Änderungen in der Konfiguration vorgenommen worden sind, schreibt diese Funktion die geänderten
    Daten in die Datenbanktabellen 'config' und 'konten' zurück
    Die geänderten Daten stehen in den Dictionarys 'd_config' und 'd_konten_names' / 'd_konten_ust'
    :param d_config:
    :param d_konten_name:
    :param d_konten_ust:
    :return: cnt = Anzahl der geänderten Datensätze
    """

    # cnt zählt die Änderungen
    cnt = 0
    for key in d_konten_name.keys():
        text = d_konten_name.get(key)
        ust = d_konten_ust.get(key)

        row = cur.execute("""SELECT bezeichnung, ustsatz FROM konten WHERE id={}""".format(key)).fetchone()
        if not row:
            # print("neuer Eintrag : key/value = ", key, ust, text)
            # Wenn eine Kontobezeichnung auf eine andere Kontoid verschoben wird, dann kann ein Fehler beim Speichern
            # auftreten, da die Kontobezeichnung auf dem alten Konto noch existiert.
            # In diesem Fall muss man in Einzelschritten ändern:
            # Zuerst die alte Bezeichnung ändern und abspeichern - und naschließend die alte Bezeichnung in einem
            # separaten Schritt neu anlegen.
            try:
                cur.execute("""INSERT INTO konten (id, ustsatz, bezeichnung) VALUES ({},{},"{}")""".
                            format(key, ust, text))
                # db.commit()
                cnt += 1
            except sq.IntegrityError:
                messagebox.showerror("Konfiguration",
                                     "Die Kontobezeichnung  '{}'  ist bereits vorhanden !\n\n".format(text) +
                                     "Bitte das Umbenennen in Einzelschritten ausführen\n" +
                                     "und jeweils einzeln speichern !\n" +
                                     "ACHTUNG: Die Änderungen wurden nicht vollständig übernommen !")

        else:
            if not row[0] == text or not row[1] == ust:
                # print("geändert:", row[0], text, row[1], ust)
                cur.execute("""UPDATE konten SET bezeichnung="{}", ustsatz="{}"  WHERE id={}""".format(text, ust, key))
                # db.commit()
                cnt += 1

    # Tabelle 'config', hier können nur Updates, aber keine Inserts erfolgen
    # dict_config["Buchungsjahr"] = "19"
    for key in d_config.keys():
        value = d_config.get(key)
        row = cur.execute("""SELECT value FROM config WHERE key="{}" """.format(key)).fetchone()
        if row:
            if not row[0] == str(value):
                cur.execute("""UPDATE config SET value="{}" WHERE key="{}" """.format(value, key))
                cnt += 1
    if cnt:
        db.commit()
    return cnt


def add_buchung(db_table, konto, date, text1, text2, netto, ust, ustsatz):
    global is_changed
    # fügt eine einzelne Buchung ein
    cur.execute("""INSERT INTO "{}" (konto, date, text1, text2, netto, ust, ustsatz)
            VALUES ({}, {}, "{}", "{}", {}, {}, {})""".format(db_table, konto, date, text1, text2, netto, ust, ustsatz))
    db.commit()
    is_changed = True


def update_buchung(db_table, key, konto, date, text1, text2, netto, ust, ustsatz):
    # überschreibt einen Datenbankeintrag
    # print("update() :", text1, netto, ust)
    global is_changed

    cur.execute("""UPDATE {} SET konto={}, date={}, text1="{}", text2="{}", netto={}, ust={}, ustsatz={}
                    WHERE id={}""".format(db_table, konto, date, text1, text2, netto, ust, ustsatz, key))
    db.commit()
    is_changed = True


def update_buchungsjahr(year):
    cur.execute("""UPDATE config SET value={} WHERE key='buchungsjahr' """.format(year))


def get_buchung(dbtable, key):
    # Holt eine einzelne Buchung, die mit dem Primary Key 'key'
    # table ist im Modul definiert und zeigt auf die angefragte Tabelle
    return cur.execute("""SELECT * FROM {} WHERE id = {}""".format(dbtable, key))


def get_data(db_table, selection='*', condition=""):
    # Holt Daten für den Treeview
    return cur.execute("""SELECT {} FROM '{}' {} ORDER BY konto, date""".format(selection, db_table, condition))


def get_konten():
    # Holt die Kontenbezeichnungen für den Treeview
    return cur.execute("""SELECT * FROM konten ORDER BY id""")


def delete_rec(db_table, key):
    cur.execute("""DELETE FROM '{}' WHERE id = {}""".format(db_table, key))
    db.commit()


def get_table_sum(db_table, _filter=""):
    # holt die Summe für netto und ust für eine ganze Tabelle
    # netto = cur.execute("""SELECT SUM(netto) FROM '{}'""".format(db_table)).fetchone()[0]
    # ust = cur.execute("""SELECT SUM(ust) FROM {}""".format(db_table)).fetchone()[0]
    row = cur.execute("""SELECT SUM(netto) As a, SUM(ust) As b FROM {} {}""".format(db_table, _filter)).fetchone()

    if row[0] is None or row[1] is None:
        row = (0, 0, 0)
    return row[0] + row[1], row[0], row[1]


# Globals für browse_table()
zeilen_pro_seite = 10
txt = None
datestring = ""
print_brutto = 0


def browse_table(db_table, db_filter="", db_label=""):
    # diese Variablen werden von anderen Funktionen genutzt / geändert
    global zeilen_pro_seite, txt, datestring, print_brutto

    date = datetime.datetime.today()
    datestring = "{:02}.{:02}.{:02}".format(date.day, date.month, date.year)
    # die Anzahl der zu druckenden Zeilen holen - bei Fehler = 15
    zeilen_pro_seite = get_config("zeilen_pro_seite", 15)

    # Ausgabe der Brutto- oder Netto-Werte in der Liste
    print_brutto = get_config("brutto_ausgabe", 0)
    txt = StringIO()

    rows = cur.execute("""SELECT * FROM {} {} ORDER BY konto, date""".format(db_table, db_filter))
    brutto_all = 0
    brutto = 0
    netto_all = 0
    netto = 0
    ust_all = 0
    ust = 0
    seite = 1
    konto = 0

    # Den Kopf der Seite 1 ausgeben
    zeile = kopfzeile(seite, db_label=db_label)
    start = True                    # Damit der erste Eintrag keine Summenzeile auslöst
    for row in rows:
        if row[KONTO] > konto:      # Die Ausgabe muss nach Konten sortiert sein, sonst funktioniert das nicht
            if not start:           # bei jedem Kontenwechsel eine Summenzeile ausgeben
                                    # Die Differenzierung nach Netto- /Bruttoausgabe wird innerhalb kontosumme()
                                    # umgesetzt
                                    # hier müssen auch die Labels "Brutto" und "Netto" getauscht werden  
                zeile = kontosumme(konto, zeile, netto, ust)

            else:                   # nur nicht am Beginn der Liste, wo das Konto von '0' auf irgendwas wechselt
                start = False

            konto = row[KONTO]      # das neue Konto merken
            brutto = 0              # Kontosummne zurücksetzen
            netto = 0
            ust = 0
            # Überschrift des Kontos in Großbuchstaben
            # aber nur, wenn noch mindestens ein weitere Zeile auf die Seite passt

            if zeile > zeilen_pro_seite - 3:
                # fusszeile berechnet brutto aus netto + ust
                fusszeile(seite, zeile, netto_all, ust_all)
                # Seitenkopf mit Seitenzahl und Zwischensummen ausgeben
                seite += 1
                zeile = kopfzeile(seite, netto_all, ust_all, db_label)

            txt.write("{:.<59} \n".format(get_name(konto).upper() + " "))
            zeile += 1

        # prüfen, ob das Seitenende erreicht ist
        if zeile > zeilen_pro_seite - 2:
            # Seitenende mit Seitenzahl und Zwischensummen ausgeben
            # Fusszeile addiert bei bedarf netto + ust zu brutto
            fusszeile(seite, zeile, netto_all, ust_all)
            seite += 1
            zeile = kopfzeile(seite, netto_all, ust_all, db_label)
            # fusszeile(seite, zeile, out, ust_all)
            # seite += 1
            # Seitenkopf mit Seitenzahl und Zwischensummen ausgeben
            # zeile = kopfzeile(seite, out, ust_all, db_label)

        # Beträge aufaddieren, dabei die Integerdarstellung durch die Division mit 100
        # auf 2 Nachkommastellen bringen (dabei entstehen Fehler in der letzten Nachkommastelle)
        n = row[NETTO] / 100
        u = row[UST] / 100
        b = n + u
        brutto_all += b
        brutto += b
        netto_all += n
        netto += n
        ust_all += u
        ust += u

        # auf 2 Nachkommastellen runden, damit es nich irgendwann zu Fehlern in den Nachkommastellen kommt
        brutto_all = round(brutto_all, 2)
        netto_all = round(netto_all, 2)
        ust_all = round(ust_all, 2)
        # print("Zeile: {}, brutto: {}, netto: {}, ust: {}".format(zeile, brutto_all, netto_all, ust_all))

        # das numerische Datum im Format 'YYMMDD' umwandeln in String "DD.MM.YY"
        date = str(row[DATE])
        date = "{}.{}.{}".format(date[4:], date[2:4], date[:2])
        if print_brutto:
            out = b
        else:
            out = n
        # Die Ausgabezeile im Währungsformat formatieren
        text = "{}  {:24} {:24}{:>11}{:>10}".format(
                date, row[TEXT1], row[TEXT2],
                lc.currency(out, symbol=False, grouping=True),
                lc.currency(u, symbol=False, grouping=True))
        txt.write("{}\n".format(text))
        zeile += 1

    # das Ende des Dokumentes ist erreicht, letzte Kontosumme drucken
    # Die Differenzierung nach Netto- /Bruttoausgabe wird innerhalb kontosumme() umgesetzt
    # hier müssen auch die Labels "Brutto" und "Netto" getauscht werden  
    zeile = kontosumme(konto, zeile, netto, ust)
    
    # bis zum Ende der Seite Leerzeilen drucken
    while zeile <= zeilen_pro_seite - 2:
        txt.write("\n")
        zeile += 1
    # der Seitenfuß der letzten Seite mit den Gesamtsuummen
    txt.write("{}\n".format('-' * 80))
    zeile += 1
    if print_brutto:
        txt.write("Gesamtsumme (Netto / Brutto / UST):{:>24}{:>11}{:>10}\n".format(
            lc.currency(netto_all, symbol=False, grouping=True),
            lc.currency(brutto_all, symbol=False, grouping=True),
            lc.currency(ust_all, symbol=False, grouping=True)))
    else:
        txt.write("Gesamtsumme (Brutto / Netto / UST):{:>24}{:>11}{:>10}\n".format(
            lc.currency(brutto_all, symbol=False, grouping=True),
            lc.currency(netto_all, symbol=False, grouping=True),
            lc.currency(ust_all, symbol=False, grouping=True)))
    return txt


def kopfzeile(seite, netto=0, ust=0, db_label=''):
    zeile = 1
    txt.write("Zusammenstellung der {:49}{}\n".format(label, datestring))
    zeile += 1
    if print_brutto:
        if seite == 1:
            txt.write("Seite:{:2}, {:54}Brutto       UST\n".format(seite, db_label))
        else:
            txt.write("Seite:{:2}, {:39}".format(seite, db_label) + "Übertrag: {:>11}{:>10}\n".format(
                lc.currency(netto + ust, symbol=False, grouping=True),
                lc.currency(ust, symbol=False, grouping=True)))
    else:
        if seite == 1:
            txt.write("Seite:{:2}, {:55}Netto       UST\n".format(seite, db_label))
        else:
            txt.write("Seite:{:2}, {:39}".format(seite, db_label) + "Übertrag: {:>11}{:>10}\n".format(
                lc.currency(netto, symbol=False, grouping=True),
                lc.currency(ust, symbol=False, grouping=True)))

    zeile += 1
    txt.write("{}\n".format('-' * 80))
    return zeile + 1


def kontosumme(konto, zeile, netto, ust):
    # wird garnichts gefunden, dann steht das Konto noch auf '0'
    if konto == 0:
        konto = "keine Einträge vorhanden"
    else:
        konto = get_name(konto)
    if print_brutto:
        txt.write("Summe  :  {:.<49}{:>11}{:>10}\n".format(
                konto + " ",
                lc.currency(netto + ust, symbol=False, grouping=True),
                lc.currency(ust, symbol=False, grouping=True)))

    else:
        txt.write("Summe  :  {:.<49}{:>11}{:>10}\n".format(
                konto + " ",
                lc.currency(netto, symbol=False, grouping=True),
                lc.currency(ust, symbol=False, grouping=True)))
    zeile += 1
    if zeile <= zeilen_pro_seite - 2:
        txt.write("\n")
        zeile += 1
    return zeile


def fusszeile(seite, zeile, netto, ust):
    while zeile <= zeilen_pro_seite - 2:
        txt.write("\n")
        zeile += 1
    txt.write("\n".format('-' * 80))
    zeile += 1
    # print("Seite: {}, Zeile: {}, netto: {}, ust: {}".format(seite, zeile, netto, ust))
    if print_brutto:
        txt.write("Seite:{:2}".format(seite) + ' ' * 26 + "Übertrag (Brutto / UST): {:>11}{:>10}\n".format(
                lc.currency(netto + ust, symbol=False, grouping=True),
                lc.currency(ust, symbol=False, grouping=True)))
    else:
        txt.write("Seite:{:2}".format(seite) + ' ' * 27 + "Übertrag (Netto / UST): {:>11}{:>10}\n".format(
                lc.currency(netto, symbol=False, grouping=True),
                lc.currency(ust, symbol=False, grouping=True)))
    # Seitenwechsel durch die Zeile <NEWPAGE> markieren
    # pdf_print() beginnt bei dieser Marke mit einer neuen Seite
    txt.write("<NEWPAGE>\n")
